#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Dec 15 23:12:29 2023

@author: anton
"""

from SLiCAP import *
from scipy.integrate import quad
from DIN_A import DinA

print("\n**********************************************************************")
print("* Assign budget for noise integration resistor                       *")
print("* and calculate feedback network drive requirements..                *")
print("**********************************************************************\n")
# Import the specifications
initSpecs = csv2specs('A1specs.csv')

i1 = instruction()

# Use a model that allows us to determine a show-stopper value of the integrator
# resistance given its relative contribution to the DIN A weighted squared output noise

fileName = "noise2"
i1.setCircuit(fileName + '.cir')
# Load circuit parameters from specifications
specs2circuit(initSpecs, i1.circuit)
i1.setSimType("numeric")

R_i                 = sp.Symbol("R_i")
f_min               = i1.getParValue('f_min')
f_max               = i1.getParValue('f_max')
V_noise             = i1.getParValue('V_onoise')
tau_i               = i1.getParValue('tau_i')
f_fp                = i1.getParValue('f_fp')
Vi_ADC              = i1.getParValue('Vi_ADC')
n_SRC               = i1.getParValue('n_SRC')
SRCnoise            = i1.getParValue('SRCnoise')

i1.setGainType("vi")
i1.setDataType("noise")
i1.setDetector("V_out")
noiseResult2 = i1.execute()
S_onoise     = noiseResult2.onoise

correction          = 1
correct             = True

n_Ri = 1

while n_Ri > 1-n_SRC or n_Ri <= 0:
    n_Ri = float(input("Assign a budget for the contribution of the noise of the integrator resistance to the DIN-A weighted squared output noise (> 0 and < %s):\n>>> "%(sp.N(1-n_SRC,3))))

N_var_required  = V_noise**2 * (n_Ri + n_SRC)

while correct:
    N_var_corrected = N_var_required * correction
    RMS_2_sq        = rmsNoise(noiseResult2, "onoise", f_min, f_max)**2
    Ri_max          = sp.solve(RMS_2_sq - N_var_corrected, R_i)[0]
    S_weighted      = S_onoise.subs(R_i, Ri_max) * DinA**2
    func_S_weighted = sp.lambdify(ini.frequency, S_weighted)
    N_var_obtained  = quad(func_S_weighted, f_min, f_max)[0]
    correction      = N_var_required/N_var_obtained
    if correction < 1.01:
        correct = False

Ri = Ri_max + 1

while Ri > Ri_max:
    Ri = float(input("Assign a value to the integrator resistance R_i (<= %s Ohm):\n>>> "%(sp.N(Ri,5))))

Ci = tau_i/Ri
ZL_min       = sp.sqrt(Ri**2 + 1/(2*sp.pi*f_fp*Ci)**2)
IL_fb_max    = sp.N(Vi_ADC/ZL_min)

S_weighted      = S_onoise.subs(R_i, Ri) * DinA**2
func_S_weighted = sp.lambdify(ini.frequency, S_weighted)
S_obtained      = quad(func_S_weighted, f_min, f_max)[0]
S_Ri_rel        = (S_obtained - SRCnoise**2)/V_noise**2

print("\nVn:  (DIN A weightd RMS output noise)                      :", sp.N(sp.sqrt(S_obtained), 3))
print("n_Ri (Relative contribution of R_i to squared output noise):", sp.N(S_Ri_rel, 3))
print("R_i  (integrator resistance)                               :", sp.N(Ri, 3))
print("C_i  (integrator capacitance)                              :", sp.N(Ci, 3))
print("I_fb (max peak-peak current in RiCi)                       :", sp.N(IL_fb_max, 3))

updatedSpecs = [spec for spec in initSpecs]

updatedSpecs.append(
        specItem('n_Ri',
        description = "Max. contribution of R_i to the DIN A weighted squared output noise",
        value       = n_Ri,
        units       = "",
        specType    = "budgets",
        )
)

updatedSpecs.append(
        specItem('Ri_max',
        description = "Max. value of integrator resistor",
        value       = Ri_max,
        units       = "Omega",
        specType    = "design",
        )
)

updatedSpecs.append(
        specItem('R_i',
        description = "Selected value of integrator resistor",
        value       = Ri,
        units       = "Omega",
        specType    = "design",
        )
)

updatedSpecs.append(
        specItem('C_i',
        description = "Required value of integrator capacitor",
        value       = Ci,
        units       = "F",
        specType    = "design",
        )
)

updatedSpecs.append(
        specItem('I_fb',
        description = "Maximum peak-peak current through feedback network",
        value       = IL_fb_max ,
        units       = "A",
        specType    = "design",
        )
)
specs2csv(updatedSpecs, 'A1specs.csv')


###############################################################################
# HTML Report
###############################################################################
htmlPage('Specifications')
head2html('Initial specifications')
text2html('The initial noise specifications are summerized in the table below.')
specs2html(initSpecs)
head2html('Updated specifications')
text2html('After finishing this design step the updated specifications are:')
specs2html(updatedSpecs)