#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Nov 27 20:31:27 2023

@author: anton
"""

from SLiCAP import *
from scipy.integrate import quad
from DIN_A import DinA

print("\n**********************************************************************")
print("* Assign a budget to the noise contribution of an NMOS input stage   *")
print("* and determine g_m and c_iss for minimum costs                      *")
print("**********************************************************************\n")

# Import the specifications
initSpecs = csv2specs('A1specs.csv')

i1 = instruction()
# Use a noise model that expresses the total noise as a function of the input
# capacitance and the transconductance of the NMOS input stage
fileName = 'noiseCissGm_n'
i1.setCircuit(fileName + '.cir')

# Load circuit parameters from specifications
specs2circuit(initSpecs, i1.circuit)

# Assign variables; get them from the specifications
f_min           = i1.getParValue('f_min')
f_max           = i1.getParValue('f_max')
V_noise         = i1.getParValue('V_onoise')
n_SRC           = i1.getParValue('n_SRC')
n_Ri            = i1.getParValue('n_Ri')

n_MN = 1

while n_MN > 1-n_SRC - n_Ri or n_MN <= 0:
    n_MN = float(input("Assign a budget for the contribution of the noise of the controller NMOS input stage to the DIN-A weighted squared output noise (> 0 and < %s):\n>>> "%(sp.N(1 - n_SRC - n_Ri , 3))))

N_var_required  = V_noise**2 * (n_MN + n_SRC + n_Ri)

# Calculate the output noise as a function of c_iss and g_m
i1.setSimType('numeric')
i1.setGainType('vi')
i1.setDataType('noise')
i1.setSource('V1')
i1.setDetector('V_out')
noiseResult = i1.execute()
S_onoise    = noiseResult.onoise
onoiseVarA  = rmsNoise(noiseResult, 'onoise', f_min, f_max)**2

###############################################################################
# Find absolute show stopper of the unweighted RMS output noise.
# this depends on the process cut-off frequency
g_m   = sp.Symbol("g_m")
c_iss = sp.Symbol("c_iss")
f_T   = sp.Symbol("f_T")
onoiseVar_fT = onoiseVarA.subs(g_m, 2*sp.pi*f_T*c_iss)
# At peak fT
onoiseVar_fTpeak = onoiseVar_fT.subs(f_T, 50E9)
# Minimum for c_iss
c_iss_best = sp.solve(sp.diff(onoiseVar_fTpeak, c_iss))[1]
# Unweighted minimum RMS noise possible:
onoiseRMS_min = sp.sqrt(onoiseVar_fTpeak.subs(c_iss, c_iss_best))
###############################################################################

DINA_correction = 1.
N_var      = DINA_correction * N_var_required
correction = True
while correction:
    onoiseVar   = assumePosParams(onoiseVarA)

    # we obtain valid combinations of g_m and c_iss by solving g_m from onoiseVar = N_var

    g_m     = sp.Symbol('g_m', positive = True)
    gm_ciss = sp.solve(onoiseVar - N_var, g_m)[0]

    # for positive values of both g_m and c_iss, we need the denominator to be positive
    # this requires c_iss to be larger than its solution from the denominator

    num, den = gm_ciss.as_numer_denom()
    ciss_min_den = sp.solve(den)[0]
    try:
        ciss_min_num = sp.solve(num)[0]
        ciss_min = np.max([ciss_min_num, ciss_min_den])
    except IndexError:
        ciss_min = ciss_min_den

    # a minimum value of g_m is found from solving the derivative of gm_ciss for
    # c_iss > ciss_min; we will do this numerically
    c_iss             = sp.Symbol('c_iss', positive = True)
    diff_gm_ciss      = sp.diff(gm_ciss, c_iss)
    func_diff_gm_ciss = sp.lambdify(c_iss, diff_gm_ciss)
    ciss_gm_min       = fsolve(func_diff_gm_ciss, float(ciss_min) + 0.1e-12)[0]
    gm_min            = gm_ciss.subs(c_iss, ciss_gm_min)

    # The product of C_iss and g_m can be taken as a measure for the costs
    # We obtain minimum cost if d(g_m*c_iss)/d(c_iss) = 0

    costs                = gm_ciss * c_iss
    diff_costs_ciss      = sp.diff(costs, c_iss)
    func_diff_costs_ciss = sp.lambdify(c_iss, diff_costs_ciss)
    ciss_costs_min       = fsolve(func_diff_costs_ciss, float(ciss_min) + 0.1e-12)[0]
    gm_costs_min         = gm_ciss.subs(c_iss, ciss_costs_min)

    #print("Minimum costs require g_m =", sp.N(gm_min, 3))
    #print("They are obtained with c_iss =", sp.N(ciss_gm_min, 3))

    # Calculate the output noise variance for minimum costs.
    ciss = ciss_costs_min
    gm  = gm_costs_min
    N_var_flat = onoiseVar.subs({c_iss: ciss, g_m: gm})

    # Plot the noise spectrum
    c_iss = clearAssumptions(c_iss)
    g_m   = clearAssumptions(g_m)
    noiseResultA = noiseResult.onoise.subs({c_iss: ciss, g_m: gm})
    f = ini.frequency
    func_OnoiseVar = sp.lambdify(f, sp.N(noiseResultA * DinA**2))
    N_var_dinA = quad(func_OnoiseVar, f_min, f_max)[0]

    # Update noise requirements after first run
    new_correction = N_var/N_var_dinA
    print("DIN A correction:", sp.N(DINA_correction, 4))
    if sp.Abs(DINA_correction - new_correction) < 0.05:
        correction = False
        noiseResult.onoise = noiseResultA
    else:
        DINA_correction = new_correction
        N_var       = DINA_correction * N_var

onoiseSpectrum = plotSweep("onoiseSpectrum_ciss_gm_N", "Output noise spectral density", noiseResult, f_min, f_max, 200, funcType="onoise", show=True)
print("RMS DIN A weighted required :", sp.N(sp.sqrt(N_var_required), 3), 'V')
print("RMS unweighted designed     :", sp.N(sp.sqrt(N_var), 3), 'V')
print("RMS unweighted obtained     :", sp.N(sp.sqrt(N_var_flat), 3), 'V')
print("RMS DIN A weighted obtained :", sp.N(sp.sqrt(N_var_dinA), 3), 'V')
print("DIN A RMS correction factor :", sp.N(sp.sqrt(N_var/N_var_dinA), 3))
print("Selected value of c_iss NMOS:", sp.N(ciss, 3), 'F')
print("Required value of g_m NMOS  :", sp.N(gm, 3), 'S')

updatedSpecs = [spec for spec in initSpecs]

updatedSpecs.append(
        specItem('c_issN',
        description = "Selected input capacitance NMOS",
        value       = ciss,
        units       = "F",
        specType    = "design",
        )
)

updatedSpecs.append(
        specItem('g_mN',
        description = "Selected transconductance input stage NMOS",
        value       = gm,
        units       = "S",
        specType    = "design",
        )
)

updatedSpecs.append(
        specItem('g_m_minN',
        description = "Minimum transconductance input stage NMOS",
        value       = gm_min,
        units       = "S",
        specType    = "design",
        )
)

updatedSpecs.append(
        specItem('c_iss_g_m_minN',
        description = "Required c_iss at g_m_min",
        value       = ciss_gm_min,
        units       = "F",
        specType    = "design",
        )
)

updatedSpecs.append(
        specItem('c_iss_costs_minN',
        description = "Required c_iss for minimum costs",
        value       = ciss_costs_min,
        units       = "F",
        specType    = "design",
        )
)

updatedSpecs.append(
        specItem('g_m_costs_minN',
        description = "g_m for minimum costs",
        value       = gm_costs_min,
        units       = "S",
        specType    = "design",
        )
)

updatedSpecs.append(
        specItem('V_onoise_gmCissN',
        description = "DIN A weighted RMS output noise with g_m, and c_iss",
        value       = sp.sqrt(N_var_dinA),
        units       = "V",
        specType    = "achievements",
        )
)
specs2csv(updatedSpecs, 'A1specs.csv')

###############################################################################
# HTML Report                                                                 #
###############################################################################

htmlPage('Circuit data NMOS')
head2html('Noise model for determination of $g_m$ and $c_{iss}$ of an NMOS')
img2html(fileName + '.svg', 1000)
netlist2html(fileName + '.cir')
elementData2html(i1.circuit)
params2html(i1.circuit)

htmlPage("Determination of input stage NMOS $g_m$ and $c_{iss}$")

head2html('Noise model for determination of $g_m$ and $c_{iss}$')
img2html(fileName + '.svg', 1000)

head2html("Determination of the output noise")
text2html("The budget for the DIN A weighted RMS noise equals $%s\\,\\mu V$."%(sp.latex(sp.N(sp.sqrt(N_var_required)*1E6, 3))))

head2html("Determination of the lowest possible unweighted RMS output noise")
text2html("The output noise spectral density in V$^2$/Hz is obtained as:")
eqn2html("S_vo", S_onoise)
text2html("Notice that this spectrum decreases with increasing $g_m$. Hence. increase of the inversion coefficient at a given $c_{iss}$ results in a lower noise.")
text2html("The unweighted squared RMS output noise $V_{onoise}^2$ over the frequency range of interest is found as:")
eqn2html("V_onoise^2", onoiseVarA)
text2html("If we substitute $g_m=2\\pi f_T c_{iss}$, and take the maximum $f_T=f_{T_{peak}}\\approx 50\\cdot 10^9$ Hz for CMOS18 NMOS, " +
          "we can calculate the minimum noise that can be obtained in this technology and for this application.")
text2html("To this end we calculate the optimum value $c_{iss_{opt}}$ for $c_{iss}$ from $\\frac{d}{d\\,c_{iss}}V_{onoise}^2 = 0$.")
text2html("Substitution of $c_{iss_{opt}}$ in the expression for $V_{onoise}^2$ " +
          "yields the minimum unweighted noise$V_{onoise_{min}}$; we find: $V_{onoise_{min}}=%s$. This is much lower than required."%(sp.latex(sp.N(onoiseRMS_min, 3))))

head2html("Determination $g_m$ and $c_{iss}$ for the best performance-to-cost ratio")
text2html("The budget for the DIN A weighted squared noise equals $%s\\, V^2$."%(sp.latex(sp.N(N_var_required, 3))))
text2html("As a result of the DIN A weighting this budget for the unweighted squeared noise be factor $%s$ more, "%(sp.latex(sp.N(DINA_correction, 3))) +
          "which equals: $%s\\, V^2$."%(sp.latex(sp.N(N_var, 3))))
text2html("This factor has been determined after a few iterations of the procedure below, " +
          "of which the results of the last iteration are shown.")
text2html("The unweighted squared RMS output noise over the frequency range of interest was found as $V_{onoise}^2$.")
text2html("We obtain valid combinations of $g_m$ and $c_{iss}$ by solving $g_m$ from:")
eqn2html("V_onoise^2", N_var)
text2html("This yields:")
eqn2html("g_m", gm_ciss)
text2html("A positive value for $g_m$ is obtained if both the numerator and the denominator of this expression are positive.")
text2html("This occurs if $c_{iss} > %s$."%(sp.latex(sp.N(ciss_min, 4))))
text2html("A minimum transconductance is required if:")
eqn2html("dg_m/dc_iss",0)
text2html("This minimum $g_{mmin}$ is found at $c_{iss}=%s$:"%(sp.latex(sp.N(ciss_gm_min,3))))
eqn2html("g_m_min", gm_min)
text2html("The product $g_m c_{iss}$ can be taken as a measure for the $costs$. The input capacitance for minimum $costs$ is obtained from:")
eqn2html("dcosts/dc_iss",0)
text2html("This yields: $c_{iss}=%s$ pF and $g_m > %s\\,\\mu $S."%(sp.latex(sp.N(ciss * 1e12,3)), sp.latex(sp.N(gm*1e6,3))))

head2html("Design verification")
text2html("The noise spectrum at the output of the receive coil amplifier is shown below.")
fig2html(onoiseSpectrum, 600)
text2html("The DIN A weighted RMS output noise over the frequency range of interest amounts $%s\\,\\mu$V."%(sp.latex(sp.N(sp.sqrt(N_var_dinA) * 1e6,3))))