#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Dec  3 21:58:15 2023

@author: anton
"""

from SLiCAP import *
from scipy.integrate import quad
from DIN_A import DinA, DINA_plot

print("\n**********************************************************************")
print("* Define PMOS finger width and length                                *")
print("* and determine the number of fingers and the drain current          *")
print("**********************************************************************\n")

fileName = 'noiseWLI_p'
# Import the specifications
specs = csv2specs('A1specs.csv')

# The channel length is free to select.
Le = 0
while Le < 0.18:
    Le = float(input("PMOS channel length in [um] (> 0.18):\n>>> "))

Le *= 1e-6

# The finger width is free to select.
W_finger = 0
while W_finger < 0.18 or W_finger > 50:
    W_finger = float(input("PMOS finger width in [um] (>= 0.18 and <=50):\n>>> "))

W_finger *= 1e-6

# Initial value of the inversion coefficient
ICinit = 1

specs.append(
        specItem('L_Mp',
        description = "Selected length input stage NMOS",
        value       = Le,
        units       = "m",
        specType    = "design",
        )
)

i1 = instruction()
i1.setCircuit(fileName + '.cir')
specs2circuit(specs, i1.circuit)

i1.setSimType('numeric')
IC_old   = i1.getParValue('IC_XU1')

correction = True
while correction:
    # Delete definitions from previous runs
    i1.delPar('IC_XU1')
    i1.delPar('W')
    i1.delPar('L')
    i1.delPar('ID')
    IC_XU1   = sp.Symbol('IC_XU1')
    L        = sp.Symbol('L')
    x_XU1    = sp.Symbol('x_XU1')

    # Get c_iss as a function of W, L and IC
    ciss_IC  = i1.getParValue('c_iss_XU1')
    ciss_W   = ciss_IC.subs({IC_XU1: ICinit, L: Le})
    ciss     = i1.getParValue('c_issP')
    W        = sp.Symbol('W')

    width    = sp.solve(ciss_W-ciss)[0]

    # round to W_finger
    Mf       = int(width/W_finger)
    W_sel    = Mf*W_finger

    gm       = i1.getParValue('g_m_XU1').subs({IC_XU1: ICinit, L: Le})
    g_m      = i1.getParValue('g_mP')
    ID       = sp.solve(gm-g_m)[0]

    # Check the inversion coefficient
    i1.defPar('IC_XU1', IC_old)
    i1.defPar('W', W_sel)
    i1.defPar('L', Le)
    i1.defPar('ID', ID)
    IC = i1.getParValue('IC_XU1')
    print("Inversion Coefficient", sp.N(IC,4))
    if sp.Abs(1-IC/ICinit) < 0.001:
        correction = False
    else:
        ICinit = IC

# design verification
f_min = i1.getParValue('f_min')
f_max = i1.getParValue('f_max')
i1.setGainType('vi')
i1.setDataType('noise')
i1.setDetector('V_out')
noiseResult    = i1.execute()
onoisePlotP    = plotSweep('NoiseVerificationP', 'Output noise spectrum PMOS input', noiseResult, f_min, f_max, 200, show=True)
RMS_unweigted  = rmsNoise(noiseResult, 'onoise', 200, 6000)
f              = ini.frequency
func_OnoiseVar = sp.lambdify(f, sp.N(noiseResult.onoise * DinA**2))
N_var_dinA     = quad(func_OnoiseVar, f_min, f_max)[0]
RMS_dinA       = np.sqrt(N_var_dinA)

print("RMS unweighted obtained     :", sp.N(RMS_unweigted, 3), 'V')
print("RMS DIN A weighted obtained :", sp.N(RMS_dinA, 3), 'V')
print("DIN A RMS correction factor :", sp.N(RMS_unweigted/RMS_dinA, 3))
print("Obtained value of c_iss     :", sp.N(i1.getParValue('c_iss_XU1'), 3), 'F')
print("Obtained value of g_m       :", sp.N(i1.getParValue('g_m_XU1'), 3), 'S')
print("Selected value of L         :", sp.N(i1.getParValue('L')*1e6, 3), 'um')
print("Obtained value of W         :", sp.N(i1.getParValue('W')*1e3, 3), 'mm')
print("Obtained value of M         :", sp.N(Mf, 3))
print("Obtained value of W_finger  :", sp.N(W_finger*1e6, 3), "um")
print("Obtained value of ID        :", sp.N(i1.getParValue('ID')*1e6, 3), 'uA')
print("Obtained value of IC        :", sp.N(i1.getParValue('IC_XU1'), 3))

updatedSpecs = [spec for spec in specs]

ID = ID
updatedSpecs.append(
        specItem('W_Mp',
        description = "Selected width input stage PMOS",
        value       = sp.N(i1.getParValue('W')),
        units       = "m",
        specType    = "design",
        )
)

updatedSpecs.append(
        specItem('g_m_Mp',
        description = "Obtained transconductance input stage PMOS",
        value       = sp.N(i1.getParValue('g_m_XU1')),
        units       = "S",
        specType    = "design",
        )
)

updatedSpecs.append(
        specItem('c_iss_Mp',
        description = "Obtained input capacitance input stage PMOS",
        value       = sp.N(i1.getParValue('c_iss_XU1')),
        units       = "F",
        specType    = "design",
        )
)

updatedSpecs.append(
        specItem('ID_Mp',
        description = "Operating current input stage PMOS",
        value       = sp.N(i1.getParValue('ID')),
        units       = "A",
        specType    = "design",
        )
)

updatedSpecs.append(
        specItem('IC_Mp',
        description = "Inversion coefficient input stage PMOS",
        value       = sp.N(i1.getParValue('IC_XU1')),
        units       = "",
        specType    = "design",
        )
)

updatedSpecs.append(
        specItem('M_Mp',
        description = "Number of fingers input stage PMOS",
        value       = Mf,
        units       = "",
        specType    = "design",
        )
)

updatedSpecs.append(
        specItem('W_F_Mp',
        description = "Finger width input stage PMOS",
        value       = W_finger,
        units       = "m",
        specType    = "design",
        )
)

updatedSpecs.append(
        specItem('V_onoise_WLI_p',
        description = "DIN A weighted RMS output noise with PMOS, W, L, and ID",
        value       = RMS_dinA,
        units       = "V",
        specType    = "achievements",
        )
)

specs2csv(updatedSpecs, 'A1specs.csv')

###############################################################################
# HTML Report                                                                 #
###############################################################################

htmlPage('Specifications')
head2html('Initial specifications for this design step')
text2html('The noise specifications at the beginning of this design step are listed below.')
specs2html(specs)
head2html('Updated specifications')
text2html('After finishing this design step the updated specifications are:')
specs2html(updatedSpecs)

htmlPage('DIN A weighting curve')
text2html('The expression for the DIN A weighting of the noise voltage is:')
eqn2html('DIN_A', DinA)
text2html('The frequency plot is shown below.')
fig2html(DINA_plot, 600)

htmlPage('Circuit data')
head2html('Noise model for determination of $W$, $L$ and $ID$.')
img2html(fileName + '.svg', 800)
netlist2html(fileName + '.cir')
elementData2html(i1.circuit)
params2html(i1.circuit)

htmlPage("Determination of $W$, $L$ and $ID$")

head2html('Noise model for determination of $W$ and $L$ and $ID$.')
img2html(fileName + '.svg', 800)

head2html("Initial estimate of the inversion coefficient and the channel length.")
text2html("We assume that this transistor with very large input capacitance and low transconductance operates in weak inversion.")
text2html("The initial value of the inversion coefficient is estamated as $%s$. This value can later be adjusted"%(sp.latex(sp.N(ICinit, 3))))
text2html("The channel length of the input transistor is selected as: $%s$."%(sp.latex(sp.N(Le, 3))))

head2html("Determination of the channel width $W$")
text2html("SLiCAP's EKV noise model 'N18_noise' evaluates the noise performance of an NMOS with $W$ and $L$ and $ID$ as input parameters.")
text2html("By doing so, it calculates the inversion coefficient from these input parameters.")
text2html("If we delete the model definition of the inversion coefficient, we can use it as input parameter, rather than the drain current.")
text2html("We can then express $c_{iss}$ in terms of the inversion coefficient:")
eqn2html("c_iss", ciss_IC)
text2html("With our first esimate of the inversion coefficient and the selected channel length, we can find the width using the result from the previous design step:")
eqn2html("c_iss", ciss)
text2html("We obtain:")
eqn2html("W", width)
text2html("We will round this to a number of $%s$ fingers, each with a width of $%s\\, \\mu$m."%(Mf, W_finger*1e6))

head2html("Determination of the drain current $ID$")
text2html("After substitution of all the know variable, the EKV noise model provides the transconductance $g_m$ as a function of the drain current $ID$:")
eqn2html("g_m", gm)
text2html("From this equation and the desired value of $g_m$ obtained from our previous design step, we obtain $ID$:")
eqn2html("ID", ID)

head2html("Design verification")
text2html("The noise spectrum at the output of the receive coil amplifier is shown below.")
fig2html(onoisePlotP, 600)
text2html("The DIN A weighted RMS output noise over the frequency range of interest amounts $%s\\,\\mu$V."%(sp.latex(sp.N(RMS_dinA * 1e6,3))))